"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const resolveVue_1 = require("../resolveVue");
const path_1 = __importDefault(require("path"));
const resolve_from_1 = __importDefault(require("resolve-from"));
const stream_1 = require("stream");
const es_module_lexer_1 = require("es-module-lexer");
const magic_string_1 = __importDefault(require("magic-string"));
const utils_1 = require("../utils");
const idToFileMap = new Map();
const fileToIdMap = new Map();
exports.modulesPlugin = ({ root, app }) => {
    // rewrite named module imports to `/__modules/:id` requests
    app.use(async (ctx, next) => {
        await next();
        if (ctx.status === 304) {
            return;
        }
        if (ctx.url === '/index.html') {
            const html = await readBody(ctx.body);
            await es_module_lexer_1.init;
            ctx.body = html.replace(/(<script\b[^>]*>)([\s\S]*?)<\/script>/gm, (_, openTag, script) => {
                return `${openTag}${rewriteImports(script)}</script>`;
            });
        }
        // we are doing the js rewrite after all other middlewares have finished;
        // this allows us to post-process javascript produced any user middlewares
        // regardless of the extension of the original files.
        if (ctx.response.is('js') &&
            // skip dependency modules
            !ctx.path.startsWith(`/__`) &&
            // only need to rewrite for <script> part in vue files
            !(ctx.path.endsWith('.vue') && ctx.query.type != null)) {
            await es_module_lexer_1.init;
            ctx.body = rewriteImports(await readBody(ctx.body));
        }
    });
    // handle /__modules/:id requests
    const moduleRE = /^\/__modules\//;
    app.use(async (ctx, next) => {
        if (!moduleRE.test(ctx.path)) {
            return next();
        }
        const id = ctx.path.replace(moduleRE, '');
        ctx.type = 'js';
        // special handling for vue's runtime.
        if (id === 'vue') {
            ctx.body = await utils_1.cachedRead(resolveVue_1.resolveVue(root).vue);
            return;
        }
        // already resolved and cached
        const cachedPath = idToFileMap.get(id);
        if (cachedPath) {
            ctx.body = await utils_1.cachedRead(cachedPath);
            return;
        }
        // source map request
        if (id.endsWith('.map')) {
            // try to reverse-infer the original file that made the sourcemap request.
            // assumes the `.js` and `.js.map` files to have the same prefix.
            const sourceMapRequest = id;
            const jsRequest = sourceMapRequest.replace(/\.map$/, '');
            const moduleId = fileToIdMap.get(jsRequest);
            if (!moduleId) {
                console.error(`[vite] failed to infer original js file for source map request ` +
                    sourceMapRequest);
                ctx.status = 404;
                return;
            }
            else {
                const modulePath = idToFileMap.get(moduleId);
                const sourceMapPath = path_1.default.join(path_1.default.dirname(modulePath), sourceMapRequest);
                idToFileMap.set(sourceMapRequest, sourceMapPath);
                ctx.type = 'application/json';
                ctx.body = await utils_1.cachedRead(sourceMapPath);
                return;
            }
        }
        // TODO support resolving from Snowpack's web_modules
        // resolve from node_modules
        try {
            const pkgPath = resolve_from_1.default(root, `${id}/package.json`);
            const pkg = require(pkgPath);
            const modulePath = path_1.default.join(path_1.default.dirname(pkgPath), pkg.module || pkg.main);
            idToFileMap.set(id, modulePath);
            fileToIdMap.set(path_1.default.basename(modulePath), id);
            ctx.body = await utils_1.cachedRead(modulePath);
        }
        catch (e) {
            console.error(e);
            ctx.status = 404;
        }
    });
};
async function readBody(stream) {
    if (stream instanceof stream_1.Readable) {
        return new Promise((resolve, reject) => {
            let res = '';
            stream
                .on('data', (chunk) => (res += chunk))
                .on('error', reject)
                .on('end', () => {
                resolve(res);
            });
        });
    }
    else {
        return stream;
    }
}
function rewriteImports(source) {
    try {
        const [imports] = es_module_lexer_1.parse(source);
        if (imports.length) {
            const s = new magic_string_1.default(source);
            let hasReplaced = false;
            imports.forEach(({ s: start, e: end, d: dynamicIndex }) => {
                const id = source.substring(start, end);
                if (dynamicIndex < 0) {
                    if (/^[^\/\.]/.test(id)) {
                        s.overwrite(start, end, `/__modules/${id}`);
                        hasReplaced = true;
                    }
                }
                else {
                    // TODO dynamic import
                }
            });
            return hasReplaced ? s.toString() : source;
        }
        return source;
    }
    catch (e) {
        console.error(`Error: module imports rewrite failed for source:\n`, source);
        return source;
    }
}
